/* GSI - general sound interface
 *
 * gsi.h
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 */

/* You can use, redistribute and/or modify this software under the terms of
 * the CGPL version 0.4 or later (see the file LICENSE.CGPL for details)
 */


#ifndef _gsi_h
#define _gsi_h

//Shared
#include <gsi/gsi_commands.h>

#ifndef NAS_SOUND
#  include <gsi/gsi_server_version.h>
#  define GSI_INTERFACE_VERSION GSI_SERVER_VERSION
#else
#  include <gsi/gsi_nas_version.h>
#  define GSI_INTERFACE_VERSION NAS_INTERFACE_VERSION
#endif

//Global
#ifdef _GSI_INTERFACE_ /* This is meant for the interface file */
int  gsi_init();
//      This initialises the GSI sound system (initialize GSI's gsi_server or connects to the NAS server when using the NAS wrapper) and is to be called before using any other GSI function.
void gsi_close();
int  gsi_init_pcm (int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile);
//      initialises the server: sounddir and soundfile may be NULL or zero length strings. In that case they are not set and no sounds will be loaded.

int  gsi_init_synth (int type);
//int  gsi_init_cd(int type);

//rename sndplay(), sndplay_stereo, sndplay_channel_stereo?
void gsi_play_sound (int slot, int command);
//      play/kill sounds using dynamically allocated dsp channels (multiple times the same sound is allowed).
//      Examples:
//      play_sound(k, SND_PLAY)  = play using new channel
//      play_sound(k, SND_REPEAT)  = ditto, repeat the sound if finished
//      play_sound(k, SND_OVERWRITE)  = overwrite LAST started k sound
//      play_sound(k, SND_KILL)  = kill LAST started k sound
//Following is not supported anymore I think (use kill_sounds() )
//      play_sound(0, SND_KILLALL)  = kill ALL sounds: NOT ALLOWED ANY MORE

void gsi_play_channel_sound (int channel, int slot, int command);
//      play a sound on a channel, other than that, equivalent to play_sound(.)

void gsi_play_stereo_sound (int slot, int command, int x, int dx);
//      same as play sound, but position the sound at x, and move it with
//      speed dx per fragment (-64 =full left, +64=full right,
//      larger values decrease volume
//      (The NAS interface ignores effects)

void gsi_play_sound_flush (int slot, int command);
//      same as play_sound but with immediate flush_sounds() appended.

int  gsi_get_channel_status(int ch);
//      status of channel ch, is there a sound playing on it?

//#define gsi_kill_sound(slot)       gsi_play_sound((slot), SND_KILL)
//#define gsi_kill_channel_sound(ch) gsi_send_commands(2, SND_CHANNEL_KILL, (ch))

void gsi_send_commands (int argc, ...);
//      send any commands to the soundserver (in byte chunks; 16bit and larger
//      must be sent LSB first)
//	eg.    general_sound(5, SND_SET_X | SND_CURRENT, x &255, x >>8, SND_SET_DX | SND_CURRENT, dx);

void gsi_flush ();
//      send all cached commands to the sound server.

void gsi_kill_sounds ();
//      equivalent to play_sound(0, SND_KILLALL). [ NOT ANY MORE ]

void set_volume (int device, int vol);
//      set volume of 'device' using a lin/log scale 0-15
//Should alter that to use internal format. Let the program do its own scaling. We use 0-255 for volume


void gsi_grab_pcm ();
//      grabs the DSP and sets the speed, channels etc. as set by earlier commands.
void gsi_grab_synth ();
//      grabs the sequencer
void gsi_grab_cd ();
//      grabs the cd

void gsi_release_pcm ();
//      releases the DSP
void gsi_release_synth ();
//      releases the sequencer
void gsi_release_cd ();
//      releases the cd


//Do vol_L/vol_R ? return value = L | (R<<8), set : (int volL, int volR)
void gsi_set_master_volume (int vol);
void gsi_set_pcm_volume (int vol);
void gsi_set_synth_volume (int vol);
void gsi_set_cd_volume (int vol);
int  gsi_get_master_volume ();
int  gsi_get_pcm_voglume ();
int  gsi_get_synth_volume ();
int  gsi_get_cd_volume ();

//Unimplemented:
void gsi_cd_init ();
void gsi_cd_stop ();
void gsi_cd_mount ();
void gsi_cd_unmount ();
void gsi_cd_eject_disk ();
void gsi_cd_start_play ();
void gsi_cd_restart_track ();
void gsi_cd_next_track ();
void gsi_cd_previous_track ();


//Low level
void _gsi_send_byte (int k);
void _gsi_send_int16 (int k);
void _gsi_send_int32 (int k);
void _gsi_send_string (char *s);
void _gsi_send_block (unsigned char /*byte*/ *data, int len);

//External
#else

extern int sound_on; //-> gsi_digi_on/gsi_midi_on/gsi_cd_on
extern int sound_volume;
//-> gsi_digi_volume/gsi_midi_volume/gsi_cd_volume
//or: gsi_sample_volume/
//gsi_sound_on/gsi_music_on/gsi_cd_on

extern int  gsi_init ();
extern void gsi_close();
extern int  gsi_init_pcm (int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile);
extern int  gsi_init_synth (int type);
//extern int  gsi_init_cd(int type);

extern void gsi_play_sound (int slot, int command);
extern void gsi_play_channel_sound (int channel, int slot, int command);
extern void gsi_play_stereo_sound (int slot, int command, int x, int dx);
extern void gsi_send_commands(int argc, ...);
extern void gsi_flush();
extern void gsi_play_sound_flush (int slot, int command);

extern int  gsi_get_channel_status(int ch); /* status of sound no, is it playing? */

#define gsi_kill_sound(slot)       gsi_play_sound((slot), SND_KILL)
#define gsi_kill_channel_sound(ch) gsi_send_commands(2, SND_CHANNEL_KILL, (ch))

extern void gsi_kill_sounds ();


//gsi_grab_dsp, gsi_release_dsp ?
extern void gsi_grab_pcm ();
extern void gsi_grab_synth ();
extern void gsi_grab_cd ();

extern void gsi_release_pcm ();
extern void gsi_release_synth ();
extern void gsi_release_cd ();

extern void gsi_set_master_volume (int vol);
extern void gsi_set_pcm_volume (int vol);
extern void gsi_set_synth_volume (int vol);
extern void gsi_set_cd_volume (int vol);
extern int  gsi_get_master_volume ();
extern int  gsi_get_pcm_volume ();
extern int  gsi_get_synth_volume ();
extern int  gsi_get_cd_volume ();

extern void gsi_send_sample (int slot, int speed, int type, int channels, int size, unsigned char *data);
//Samples must in MSB format (for 16 bit sounds), and left channel first (for stereo sounds)
extern int  gsi_load_file (int slot, char *filename);

//To make up more sophisticated effects (eg stereo position adjustment, observer rotation) send commands yourself to the server:
//#define gsi_send_commands      general_sound
//You can also use:
#define gsi_send_byte         _gsi_send_byte
#define gsi_send_int16        _gsi_send_int16
#define gsi_send_int32        _gsi_send_int32
#define gsi_send_string       _gsi_send_string
#define gsi_send_block        _gsi_send_block 


//Unimplemented:
extern void gsi_cd_init ();
extern void gsi_cd_stop ();
extern void gsi_cd_mount ();
extern void gsi_cd_unmount ();
extern void gsi_cd_eject_disk ();
extern void gsi_cd_start_play ();
extern void gsi_cd_restart_track ();
extern void gsi_cd_next_track ();
extern void gsi_cd_previous_track ();
#endif /* INTERFACE / ! INTERFACE */


#endif /* _gsi_h */
