/* GSI - general sound interface
 *
 * gsi_interface.c
 *
 * Copyright (C) W.H.Scholten 1996,1997,1998
 *
 * ** See gsi.h for a description of the commands **
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#ifdef __STDC__
#  include <stdlib.h>
#endif
#include <string.h>

#include <stdarg.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>


//Following needed for mixer (& dma)
#ifdef __linux__
#  include <linux/soundcard.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
#endif

#define  _GSI_INTERFACE_
#include <gsi/gsi.h>


#include <gsi/mytypes.h> /* use gsi_int16 etc instead ? */
//#define DEBUG
#include <gsi/debug.h>

/*__________________VERSIE___________________*/


/*__________________VARIABELEN___________________*/
char *mixerdev=MIXERDEV;
int  sound_volume=80; //default PCM volume
int  sound_on=1;

static int  soundfd;
//FEEDBACK
static int  statusfd;

static char sndserver_ok = 1;
static char mixer_ok = 1;

static int  num_commands;
static byte gsi_command_buffer[256];

static int  GSI_master_vol;
static int  GSI_pcm_vol;
static int  GSI_synth_vol;
static int  GSI_cd_vol;

static int  server_id; /* process id of the gsi_server */

int  channel_status[8]; /* 256 bits (playing yes/no for each channel), i.e. an int must be 32 bit or more */

/*__________________DEFINITIES___________________*/
#define LINEAR_VOLUME

#define MAX_QUEUE_LENGTH 256 /* sound commands stored */

#define SOUND_BIN_DIR "/usr/local/bin"
#define SOUNDSERVER "gsi_server"

/*__________________DATA___________________*/
static int volume_table[16]={
#ifdef LOG_VOLUME
	0,30,45,55,67,70,73,76,79,82,85,88,91,94,97,100
#endif
#ifdef LINEAR_VOLUME
	0,7,13,20,27,33,40,47,53,60,67,73,80,87,93,100
#endif
};



//Low level
void _gsi_send_byte(int k);
void _gsi_send_int16(int k);
void _gsi_send_int32(int k);
void _gsi_send_string(char *s);
void _gsi_send_block(byte *data, int len);

/*__________________PROGRAMMA___________________*/

void child_killed(int i) {
	int status;

//	printf("child killed, i=%d\n", i);
//	wait(&status); //wait for the child to exit

//	if (kill(server_id, 0)) //send a NOP, fails if our sound process is dead
	if (wait(&status) == server_id)
		{
			sndserver_ok = 0;
			close(soundfd);
//FEEDBACK
			close(statusfd);
			fprintf(stderr, "sound server killed\n");
		}
}



/////////// Set sound volume /////////////
/* Mixer volume settings:
 *   VOLUME : master volume, i.e. headphone and line out.
 *   PCM    : dsp volume
 *   SYNTH  : onboard midi playing device
 */

/* saving volumes is done to remove any rounding problems with repeated reading/writing (device values read may not be the same as what has been set) */
/* Should add 'volume_set' flags so that recall before set will return the real hardware volume */
void gsi_set_master_volume(int vol) {
	GSI_master_vol=vol; set_volume(SOUND_MIXER_WRITE_VOLUME, vol);
}
void gsi_set_pcm_volume(int vol) {
	GSI_pcm_vol=vol; set_volume(SOUND_MIXER_WRITE_PCM, vol);
}
void gsi_set_synth_volume(int vol) {
	GSI_synth_vol=vol; set_volume(SOUND_MIXER_WRITE_SYNTH, vol);
}
void gsi_set_cd_volume(int vol) {
	GSI_cd_vol=vol; set_volume(SOUND_MIXER_WRITE_CD, vol);
}

int gsi_get_master_volume() { /* if (GSI_master_vol<0) GSI_master_vol=read_volume(SOUND_MIXER_READ_VOLUME);*/ return GSI_master_vol; }
int gsi_get_pcm_volume() { /* if (GSI_pcm_vol<0) GSI_pcm_vol=read_volume(SOUND_MIXER_READ_PCM);*/ return GSI_pcm_vol; }
int gsi_get_synth_volume() { /* if (GSI_synth_vol<0) GSI_synth_vol=read_volume(SOUND_MIXER_READ_SYNTH);*/ return GSI_synth_vol; }
int gsi_get_cd_volume() { /* if (GSI_cd_vol<0) GSI_cd_vol=read_volume(SOUND_MIXER_READ_CD);*/ return GSI_cd_vol; }

/* We expect volumes in the range 0-255 */
void set_volume(int device, int vol) {
	int mixer, volume;

	if (mixer_ok) {
		mixer=open(mixerdev, O_WRONLY | O_NONBLOCK);
		sound_volume = (vol & 65535);
		if (sound_volume >255) sound_volume=255;
//		volume=volume_table[sound_volume];
		volume=(sound_volume*100)/255; /* OSS uses a scale 0-100 */
DEB1(fprintf(stderr, "set volume: sound_volume=%d -> %d real\n", sound_volume, volume));
		volume |=volume<<8; /* Left and right channels */
		ioctl(mixer, device, &volume);
		close(mixer);
	}
}

int gsi_init() {
		//here read the hardware master volume, pcm volume, synth volume cd volume
		//Or do this with a different piece of code above? (init volumes at -1)
	return 0;
}

//init_dsp ?
int  gsi_init_pcm(int dsp_speed, int dsp_channels, int dsp_bytes, int dsp_fragsize, char *sounddir, char *soundfile) {
	int  child, fd[2];
//FEEDBACK
	int  fd2[2];

	char *argv[3];
//   char sndserver_filename[512];
	char *sndserver_filename;

	if (!(sndserver_filename=getenv("GSI_SERVER"))) sndserver_filename=SOUND_BIN_DIR"/"SOUNDSERVER;
//   strcpy(sndserver_filename, BINDIR"/"SOUNDSERVER);
	signal(SIGCHLD, &child_killed);

		/* should set SIGPIPE handler too */
//	signal(SIGPIPE, SIG_IGN);

		// Create a pipe, set the write end to close when we exec the sound server,
		// and set both ends to non-blocking (in the server)

	pipe(fd);
	soundfd=fd[1]; /* writeable pipe */

//FEEDBACK	(readable pipe):
//FEEDBACK
	pipe(fd2);
//FEEDBACK
	statusfd=fd2[0];

	child=fork();

	if( child == 0 )  { /* child's thread */
		close(fd[1]);
//FEEDBACK
		close(fd2[0]);

		dup2(fd[0], STDIN_FILENO);
		close(fd[0]);
//FEEDBACK
		dup2(fd2[1], STDOUT_FILENO);
//FEEDBACK
		close(fd2[1]);

		argv[0] = sndserver_filename;
		argv[1] = "-loop";
		argv[2] = NULL;
		execvp(sndserver_filename, argv);
			//We only arrive here if execvp failed to replace this process with the sounserver code
		fprintf (stderr, "Couldn't execute %s!\n", sndserver_filename);
		exit (0); /* child quits */
	}

	close(fd[0]);
//FEEDBACK
	close(fd2[1]);

	server_id=child;

#if 0 /* not needed, SIGCHLD will do this */
	if (kill(server_id, 0)) //send a NOP, fails if the process is dead
		{
			sndserver_ok = 0;
			close(soundfd);
//FEEDBACK
			close(statusfd);
		}
#endif
		/* should wait for ok from the read-pipe */

		/* Call the plumber to unblock the 'in' pipe! */
//	fcntl(statusfd, F_SETFL, O_NONBLOCK);
	fcntl(fd2[0], F_SETFL, O_NONBLOCK);

	DEB1(fprintf(stderr, "SNDI: sndserver_ok=%d\n", sndserver_ok));

	if (sndserver_ok) {
			//server starts PAUSE'd, DSP released.
		_gsi_send_byte(SND_SET_DSP_SPEED);   _gsi_send_int16(dsp_speed);
		DEB1(fprintf(stderr, "SNDI: set speed %d\n", dsp_speed));
		_gsi_send_byte(SND_SET_CHANNELS);   _gsi_send_byte(dsp_channels);
		DEB1(fprintf(stderr, "SNDI: set chan  %d\n", dsp_channels));
		_gsi_send_byte(SND_SET_BYTES);      _gsi_send_byte(dsp_bytes);
		DEB1(fprintf(stderr, "SNDI: set bytes %d\n", dsp_bytes));
		_gsi_send_byte(SND_SET_FRAGSIZE);   _gsi_send_byte(dsp_fragsize);
		DEB1(fprintf(stderr, "SNDI: set frag  %d\n", dsp_fragsize));
		_gsi_send_byte(SND_SET_SOUND_DIR);   _gsi_send_string(sounddir);
		DEB1(fprintf(stderr, "SNDI: set sounddir  %s\n", sounddir));
		_gsi_send_byte(SND_LOAD_SOUND_LIST); _gsi_send_string(soundfile);
		DEB1(fprintf(stderr, "SNDI: set soundsfile  %s\n", soundfile));
		_gsi_send_byte(SND_LOAD_LISTED);
		DEB1(fprintf(stderr, "SNDI: SND_LOAD_LISTED\n"));
		_gsi_send_byte(SND_GRAB_DSP); //This sets the new parameters.
		DEB1(fprintf(stderr, "SNDI: SND_GRAB_DSP\n"));
	}

	{
	int mixer, value;
	mixer = open(mixerdev,O_WRONLY | O_NONBLOCK);
	if(mixer==-1) {
		mixer_ok=0;
		fprintf(stderr,"sound: Couldn't open mixer %s\n", mixerdev);
	}
	else {
//		set_master_volume(sound_volume);
		gsi_set_pcm_volume(sound_volume);
		close(mixer);
	}
	}
	fprintf(stderr, "Sound server initialized.\n");
}

void gsi_kill_sounds() {
	num_commands=0;
	_gsi_send_byte(SND_KILLALL);
}


//Send a list of commands with data (bytes):
void gsi_send_commands(int argc, ...) {
	va_list argp;

	va_start(argp, argc);
	while (argc--) {
		gsi_command_buffer[num_commands++]=va_arg(argp, byte);
		if (num_commands==MAX_QUEUE_LENGTH) gsi_flush();
	}
}


void gsi_play_stereo_sound (int slot, int command, int x, int dx) {
DEB1(fprintf(stderr, "SNDI: play_stereo_sound\n"));
	if (sound_on && sndserver_ok) {
		gsi_send_commands(7, command, slot, SND_SET_X | SND_CURRENT, x &255, x >>8, SND_SET_DX | SND_CURRENT, dx);
	}
}

void gsi_play_sound (int slot, int command) {
	if (sound_on && sndserver_ok) {
		if (num_commands>=MAX_QUEUE_LENGTH-1) gsi_flush();
		DEB1(fprintf(stderr, "SNDI: play_sound (slot %d, cmd %d), num_commands now %d\n", slot, command, num_commands+2));

		gsi_command_buffer[num_commands++]=command;
		gsi_command_buffer[num_commands++]=slot;
	}
}

void gsi_play_channel_sound (int channel, int slot, int command) {
	if (sound_on && sndserver_ok) {
		if (num_commands>=MAX_QUEUE_LENGTH-2) gsi_flush();
		gsi_command_buffer[num_commands++]=command | SND_CHANNEL;
//		gsi_command_buffer[num_commands++]=command;
		gsi_command_buffer[num_commands++]=channel;
		gsi_command_buffer[num_commands++]=slot;
	}
}

static byte status_queue[4];
static int  status_queue_len;
void _gsi_check_callback() {
	byte b;
	int  i;
	int  decoded_length=0;

DEB1(fprt(stderr, "SNDI: callback check"));
	for(;;) {
		i=read(statusfd, &b, 1);
DEB1(fprt(stderr, "SNDI: callback check: read status=%d\n", i));
		if(i==0) break; //EOF on pipe, i.e. child died. Just quit and let the handler disable sound.
DEB1(if (i<0) fprintf(stderr, "SNDI: callback errno=%d\n", errno));
		if (i<0) break; //nothing
		if (i>0) { //there was something
DEB1(fprintf(stderr, "SNDI: callback\n"));
			status_queue[status_queue_len++]=b;
//			int command=b; //=> if (command_size==0) command=b;
			switch(status_queue[0]) {
			case GSI_CALLBACK_OK:
				decoded_length=1;
				break;
			case GSI_CALLBACK_ERROR: /* general error */
				decoded_length=1;
				break;
			case GSI_CALLBACK_NO_SUCH_FILE:
				decoded_length=1;
				fprintf(stderr, "no such file\n");
				break;
			case GSI_CALLBACK_BAD_LIST: /* eg sounds that can't be loaded ?? */
			case GSI_CALLBACK_BAD_COMMAND:
			case GSI_CALLBACK_DSP_BUSY:
			case GSI_CALLBACK_SYNTH_BUSY:
				decoded_length=1;
				break;
//			case GSI_CALLBACK_END: //useless for stack, dump this (or add sample id's generated when starting a sound, but must use callbacks etc. Not worth it..
//					/* <slot> 1 1 number of the sound that finished */
//				break;
			case GSI_CALLBACK_CHANNEL_END:
					/* <ch> 1 1 */
			{
				int  j;
				if (status_queue_len<2) continue;
				decoded_length=2;
				j=status_queue[1];
				channel_status[j>>5] &=~(1<<(j &31));
				DEB1(fprt(stderr, "GSI_CALLBACK_CHANNEL_END: ch=%d\n", j));
			}
			break;
#if 1
/* why is this needed? Suppose a sound is killed or ends, a play_channel_sound is given and gsi_set_channel_status(ch, 1) given, then a kill CALLBACK can come after this and channel status will be 0 even though the sound has just started. This means we need either a counter (increase with gsi_adjust_channel_status(ch) which increses it by 1, CALLBACK decreases it by 1, or callbacks (which will arrive in the same order as things change in the server.
KILLALL must give a callback to reset all counters to 0
 */
			case GSI_CALLBACK_CHANNEL_STARTED:
					/* <ch> 1 1 */
			{
				int  j;
				if (status_queue_len<2) continue;
				decoded_length=2;
				j=status_queue[1];
				channel_status[j>>5] |=(1<<(j&31));
				DEB1(fprt(stderr, "GSI_CALLBACK_CHANNEL_STARTED: ch=%d\n", j));
			}
			break;
			case GSI_CALLBACK_CHANNEL_ENDALL:
				decoded_length=1;
			{
				int  j;
				for (j=0; j<8; j++) channel_status[j] = 0;
				DEB1(fprt(stderr, "GSI_CALLBACK_CHANNEL_ENDALL\n"));
			}
			break;
#endif
			default: fprintf(stderr, "SNDI: Unkown callback %d\n", status_queue[0]);
				decoded_length=1; /* remove the byte */
			}

			if (decoded_length>0) {
				status_queue_len -=decoded_length;
				if (status_queue_len>0) memmove(status_queue, status_queue+decoded_length, status_queue_len);
			}
		}
	}
}

//This functions is useless for stack playing, scrap it.
int  gsi_sound_status(int slot) {
//	gsi_check_callback(); //handle any data on the input pipe
	return 0; /* fake result that it is not playing */
}
int  gsi_get_channel_status(int ch) {
	_gsi_check_callback(); //handle any data on the input pipe
	return channel_status[ch >>5] & (1<< (ch&31));
//	return 0; /* fake result that it is not playing */
}

//void gsi_check_callback() {
//
//}

void gsi_close() {
	gsi_release_pcm(); //also release_cd/release_synth
	if (sndserver_ok) {
		_gsi_send_byte(SND_CLOSE_SERVER);
	}
	sndserver_ok=0;
}


void gsi_flush() {
	int i;

//	for (i=0; i<num_commands; i++) {
//		_gsi_send_byte(gsi_command_buffer[i]);
//	}
	if (num_commands>0)
		_gsi_send_block(&gsi_command_buffer[0], num_commands);
	num_commands=0;
DEB1(fprintf(stderr, "goto callback check\n"));
_gsi_check_callback();
}

void gsi_play_sound_flush (int slot, int command) {
	gsi_play_sound (slot, command);
	gsi_flush();
}

//separate pause_pcm?
//pause all (pcm/synth/cd)
void gsi_pause() {
	if (sndserver_ok) _gsi_send_byte(SND_PAUSE_ALL);
}

//continue all (pcm/synth/cd)
void gsi_continue() {
	if (sndserver_ok) _gsi_send_byte(SND_CONTINUE_ALL);
}

//add gsi_grab_devices() ?
void gsi_grab_pcm() {
	if (sndserver_ok) _gsi_send_byte(SND_GRAB_DSP);
}


//add gsi_release_devices() ?
void gsi_release_pcm() {
	if (sndserver_ok) _gsi_send_byte(SND_RELEASE_DSP);
}


//Data & file stuff
int  gsi_load_file(int slot, char *filename)
{
	_gsi_send_byte(SND_LOAD_SOUND_FILE);
	_gsi_send_byte(slot);
	_gsi_send_string(filename);
	return 0;
}

void gsi_send_sample(int slot, int speed, int type, int channels, int size, unsigned char *data) {
	int  i;

		/* This gets sent in sync mode */
	_gsi_send_byte(SND_LOAD_SOUND_DATA);
	_gsi_send_byte(slot);
	_gsi_send_int32(speed);
	_gsi_send_byte(type);
	_gsi_send_byte(channels);
	_gsi_send_int32(size);
	_gsi_send_block(data, size);
}


///////////// Pipe functions (internal use only) ///////////////
void _gsi_send_block(byte *data, int len) {
	fd_set write_fd_sel;
	int i=0, j=0;

	DEB1(fprt(stderr, "SNDI: _gsi_send_block: sending a block of %d bytes\n", len));
	if (sndserver_ok) {
		while((j=write (soundfd, data+i, len-i)) <len-i) {
			if (j>0) i +=j;
			FD_ZERO(&write_fd_sel);
			FD_SET(soundfd, &write_fd_sel);
			select(soundfd+1, NULL, &write_fd_sel, NULL, NULL);
		}
	}
}
void _gsi_send_byte(int data) {
	fd_set write_fd_sel;
	byte pipe_byte=data &255;
	DEB2(fprt(stderr, "SNDI: _gsi_send_byte: sending %d\n", pipe_byte));
	if (sndserver_ok) {
		while(write (soundfd, &pipe_byte, sizeof (byte)) <1) {
			FD_ZERO(&write_fd_sel);
			FD_SET(soundfd, &write_fd_sel);
			select(soundfd+1, NULL, &write_fd_sel, NULL, NULL);
		}
	}
}
void _gsi_send_int16(int data) {
	if (sndserver_ok) {
		_gsi_send_byte(data);
		_gsi_send_byte(data >>8);
	}
}
void _gsi_send_int32(int data) {
	if (sndserver_ok) {
		_gsi_send_byte(data);
		_gsi_send_byte(data >>8);
		_gsi_send_byte(data >>16);
		_gsi_send_byte(data >>24);
	}
}
void _gsi_send_string(char *string) {
//	int n;
	if (sndserver_ok) {
		if (string==NULL)
			_gsi_send_byte(0);
		else
			_gsi_send_block(string, strlen(string)+1);
//			n=write (soundfd, string, strlen(string)+1);
	}
	DEB1(fprintf(stderr, "SNDI: string length=%d\n", string==NULL ? 0 : strlen(string)));
}
